/* SPDX-License-Identifier: GPL-2.0 */

#ifndef _ASM_AVR_SYSTEM_H
#define _ASM_AVR_SYSTEM_H

#include <asm/io.h>

typedef unsigned char irqflags_t;

#define local_irq_disable()     __asm__ __volatile__("cli": : :"memory")
#define local_irq_enable()      __asm__ __volatile__("sei": : :"memory")

#define local_save_flags(x)  do { x = in(SREG); } while(0)
#define local_irq_save(x)    do { x = in(SREG); local_irq_disable(); } while(0)
#define local_irq_restore(x) do { out(SREG, x); } while(0)

#define irqs_disabled_flags(flags) (!(flags & SREG_I))
#define irqs_disabled()            (!(in(SREG) & SREG_I))


#ifdef SMCR
#define SMCR_         SMCR
#define SMCR_SE_      SMCR_SE
#define SMCR_SM_IDLE_ SMCR_SM_IDLE
#elif defined(MCUCR_SM_IDLE)
#define SMCR_         MCUCR
#define SMCR_SE_      MCUCR_SE
#define SMCR_SM_IDLE_ MCUCR_SM_IDLE
#endif

#ifdef SMCR_

#define set_sleep_mode(sm)  out(SMCR_, in(SMCR_) | sm)

#define do_sleep() do { \
	out(SMCR_, in(SMCR_) | SMCR_SE_); \
	asm volatile("sleep"); \
	out(SMCR_, in(SMCR_) & ~SMCR_SE_); \
} while(0);

#else

#define set_sleep_mode(sm) do{}while(0)
#define do_sleep()         do{}while(0)

#endif

#ifdef CONFIG_WAKE_SEMA_FROM_IRQ	

#define safe_halt()    do { \
	local_irq_enable(); \
	set_sleep_mode(SMCR_SM_IDLE_); \
	do_sleep(); \
} while(0)

#else

#define safe_halt()           do { \
	local_irq_enable(); \
	for(;;) { \
		set_sleep_mode(SMCR_SM_IDLE_); \
		do_sleep(); \
	} \
} while(0)

#endif

#endif /* __ASM_SYSTEM_H__ */
